/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


static void * WriteInvalid(DWORD dwAddress)
{
	DPF(DEBUG_MEMWARN, "Illegal Memory Access Tried to Write To 0x%08x PC: 0x%08x", dwAddress, g_dwPC);
	if (g_bWarnMemoryErrors)
	{
		CPU_Halt("Illegal Memory Access");
		DBGConsole_Msg(0, "Illegal Memory Access: Tried to Write To 0x%08x (PC: 0x%08x)", dwAddress, g_dwPC);
	}


	return g_pMemoryBuffers[MEM_UNUSED];

}

static void *Write_Noise(DWORD dwAddress)
{
	static BOOL bWarned = FALSE;
	//if (!bWarned)
	{
	//	DBGConsole_Msg(0, "Writing noise (0x%08x) - sizing memory?", dwAddress);
		bWarned = TRUE;
	}
	return g_pMemoryBuffers[MEM_UNUSED];

}

static void * WriteMappedGolden(DWORD dwAddress)
{
	//DWORD dwTranslated = dwGEOff;//(dwAddress- 0x7f000000) + GOLDENEYE_HACKED_OFFSET;
	//DWORD dwTranslated = dwAddress + GOLDENEYE_HACKED_OFFSET;
	//DBGConsole_Msg(0, "Writing Golden Mapped from 0x%08x, 0x%08x", dwAddress, dwTranslated);
	//return g_pMemoryBuffers[MEM_UNUSED];


	return Write_Noise(dwAddress);

}


static void * WriteMapped(DWORD dwAddress)
{
	DWORD i;
	BOOL bMatched;
	DWORD iMatched;
	static DWORD iLastMatched = 0;

	g_dwTLBWHit++;
	//DBGConsole_Msg(0, "WriteMapped");

	// Hack to check most recently reference entry first
	if ((dwAddress & g_TLBs[iLastMatched].vpnmask) == (g_TLBs[iLastMatched].addrcheck))
	{
		iMatched = iLastMatched;
		goto matched;
	}

	bMatched = FALSE;
	for (i = 0; i < 32; i++)
	{
		struct TLBEntry & tlb = g_TLBs[i];

		// Check that the VPN numbers match
		if ((dwAddress & tlb.vpnmask) == tlb.addrcheck)
		{
			bMatched = TRUE;
			iMatched = i;
			iLastMatched = i;

			// touch entry to show it has been referenced recently
			// we don't need to do this if this was refenced previously
			g_TLBs[iMatched].lastaccessed = ++s_nTLBTimeNow;
			break;
		}
	}


	if (!bMatched)
	{
		// TLBRefill

		// No valid TLB entry - throw TLB Refill Exception
		TLB_Refill_Store(dwAddress);
		return g_pMemoryBuffers[MEM_UNUSED];
	}

matched:

	//DBGConsole_Msg(0, "ReadMapped (%d)", iMatched);
	struct TLBEntry & tlb = g_TLBs[iMatched];

	
	// Matched entry
	if (!tlb.g)
	{
		// Do ASID stuff
		CPU_Halt("TLB: ASID Stuff (Not Implemented)");
		DPF(DEBUG_TLB, "TLB: ASID Stuff (NoImpl) (Index %d)", i);
		return g_pMemoryBuffers[MEM_UNUSED];
	}

	// Global
	if (dwAddress & tlb.checkbit) // Odd entry
	{
		if (tlb.pfno & TLBLO_V)
		{
			// Valid
			DWORD dwAddr = tlb.pfnohi | (dwAddress & tlb.mask2);

			return (g_pu8RamBase + (dwAddr & 0x007FFFFF));
		}
		else
		{
			// Throw TLB Invalid exception
			TLB_Invalid_Store(dwAddress);
			return g_pMemoryBuffers[MEM_UNUSED];
		}
	}
	else   // Even entry 
	{
		if (tlb.pfne & TLBLO_V)
		{
			// Valid
			DWORD dwAddr = tlb.pfnehi | (dwAddress & tlb.mask2);

			return (g_pu8RamBase + (dwAddr & 0x007FFFFF));
		}
		else
		{
			// Throw TLB Invalid exception
			TLB_Invalid_Store(dwAddress);
			return g_pMemoryBuffers[MEM_UNUSED];
		}
	}

}



static void *Write_RAM_4Mb_8000_803F(DWORD dwAddress)
{
	//DPF(DEBUG_MEMORY, "Writing to RDRAM: 0x%08x", dwAddress);
	return g_pu8RamBase + (dwAddress & 0x003FFFFF);
}

static void *Write_RAM_8Mb_8000_807F(DWORD dwAddress)
{
	//DPF(DEBUG_MEMORY, "Writing to RDRAM: 0x%08x", dwAddress);
	return g_pu8RamBase + (dwAddress & 0x007FFFFF);
}



// 0x03F0 0000 to 0x03FF FFFF  RDRAM registers
static void *Write_83F0_83F0(DWORD dwAddress)
{
	
	if ((dwAddress&0x1FFFFFFF) < 0x04000000)
	{
		DPF(DEBUG_MEMORY_RDRAM_REG, "Writing to MEM_RD_REG: 0x%08x", dwAddress);
	//	DBGConsole_Msg(0, "Writing to MEM_RD_REG: 0x%08x", dwAddress);

		return ((BYTE *)g_pMemoryBuffers[MEM_RD_REG] + (dwAddress & 0xFF));
	}
	else
	{
		return WriteInvalid(dwAddress);
	} 
}

// 0x0400 0000 to 0x0400 FFFF  SP memory
static void * Write_8400_8400(DWORD dwAddress)
{

	if ((dwAddress&0x1FFFFFFF) <= SP_IMEM_END)
	{
		DPF(DEBUG_MEMORY_SP_IMEM, "Writing to SP_MEM: 0x%08x", dwAddress);

		return ((BYTE *)g_pMemoryBuffers[MEM_SP_MEM] + (dwAddress & 0x1FFF));
	}
	else
	{	
		return WriteInvalid(dwAddress);
	}
}

